""" HTML string parts for HTML export of .eln files """


htmlStart = """
<!DOCTYPE HTML>
<html lang='en'>
<head>
  <meta http-equiv='Content-type' content='text/html;charset=UTF-8' />
  <meta name='viewport' content='width=device-width, initial-scale=1'>
  <meta name='description' content='This is a .eln export from PASTA-ELN'>
  <title>ELN archive HTML preview - PASTA-ELN</title>
  <style>
    body {font-family: 'sans-serif';}
    main {contain: paint;}
    table {
      width: 100%;
      border-collapse: collapse;
      font-family: system-ui, sans-serif;
      margin: 1em 0;
      box-shadow: 0 2px 4px rgba(0, 0, 0, 0.05);
    }
    th, td {
      padding: 0.75rem 1rem;
      text-align: left;
      border-bottom: 1px solid #e2e2e2;
    }
    tbody tr:nth-child(even) {background-color: #c7c7c7;}
    tbody tr:hover {background-color: #afc8ff;}
    th {
      font-weight: 600;
      letter-spacing: 0.5px;
    }
    details {
      padding: 0.5rem;
      border-radius: 0.5rem;
    }
  </style>
</head>
<body>
  <main>
"""

htmlEnd = """
  </main>
  <hr>
  <footer>ELN HTML preview - Generated by PASTA-ELN version ___VERSION___ on ___DATE___</footer>
</body>

</html>
"""

importantKeys = ['name', 'description', 'keywords','hasPart', 'value', 'unitText', 'identifier']
